/*
 * Copyright (C) 2006-2007 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file dccp_send.c paraslash's dccp sender */

/*
 * based on server.c of dccp-cs-0.01.tar.bz2,
 * (C) 2005 Ian McDonald <imcdnzl@gmail.com>
 */

#include "server.h"
#include "net.h"
#include "list.h"
#include "vss.h"
#include "send.h"
#include "dccp.h"
#include "error.h"
#include "string.h"
#include "fd.h"
#include "close_on_fork.h"
#include "chunk_queue.h"
#include "server.cmdline.h"

/** the list of connected clients **/
static struct list_head clients;
static int listen_fd = -1;
static struct sender *self;

/** Maximal number of bytes in a chunk queue. */
#define DCCP_MAX_PENDING_BYTES 40000

/** describes one connected client */
struct dccp_client {
	/** the dccp socket */
	int fd;
	/** address information about the client */
	struct sockaddr_in addr;
	/** the position of this client in the client list */
	struct list_head node;
	/** non-zero if audio file header has been sent */
	int header_sent;
	/** The list of pending chunks for this client. */
	struct chunk_queue *cq;
};

static void dccp_pre_select( int *max_fileno, fd_set *rfds,
		__a_unused fd_set *wfds)
{
	if (listen_fd < 0)
		return;
	FD_SET(listen_fd, rfds);
	*max_fileno = PARA_MAX(*max_fileno, listen_fd);
}

static void dccp_post_select(fd_set *rfds, __a_unused fd_set *wfds)
{
	struct dccp_client *dc;
	int ret;

	if (!FD_ISSET(listen_fd, rfds))
		return;
	dc = para_calloc(sizeof(struct dccp_client));
	ret = para_accept(listen_fd, &dc->addr, sizeof(struct sockaddr_in));
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
		return;
	}
	PARA_NOTICE_LOG("connection from %s\n", inet_ntoa(dc->addr.sin_addr));
	dc->fd = ret;
	para_list_add(&dc->node, &clients);
	add_close_on_fork_list(dc->fd);
	mark_fd_nonblock(dc->fd);
	dc->cq = cq_new(DCCP_MAX_PENDING_BYTES);
}

static int dccp_open(void)
{
	struct sockaddr_in servaddr;
	int ret;

	ret = dccp_get_socket();
	if (ret < 0)
		return ret;
	listen_fd = ret;

	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servaddr.sin_port = htons(conf.dccp_port_arg);
	ret = bind(listen_fd, (struct sockaddr *)&servaddr, sizeof(servaddr));
	if (ret < 0)
		return -E_DCCP_BIND;
	ret = dccp_set_socket(listen_fd);
	if (ret < 0)
		return ret;
	ret = listen(listen_fd, 0);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", strerror(errno));
		return -E_DCCP_LISTEN;
	}
	PARA_DEBUG_LOG("listening on fd %d\n", listen_fd);
	add_close_on_fork_list(listen_fd);
	mark_fd_nonblock(listen_fd);
	return 1;
}

static void dccp_shutdown_client(struct dccp_client *dc)
{
	PARA_DEBUG_LOG("shutting down %s (fd %d)\n", inet_ntoa(dc->addr.sin_addr),
		dc->fd);
	close(dc->fd);
	del_close_on_fork_list(dc->fd);
	cq_destroy(dc->cq);
	list_del(&dc->node);
	free(dc);
}

/*
 * ret: Negative on errors, zero if nothing was written and write returned
 * EAGAIN, number of bytes written else.
 */
static int dccp_write(int fd, const char *buf, size_t len)
{
	size_t written = 0;
	int ret = 0;

	while (written < len) {
		ret = write(fd, buf + written, PARA_MIN(1024, len - written));
		/*
		 * Error handling: CCID3 has a sending wait queue which fills up and is
		 * emptied asynchronously. The EAGAIN case means that there is currently
		 * no space in the wait queue, but this can change at any moment and is
		 * thus not an error condition.
		 */
		if (ret < 0 && errno == EAGAIN)
			return written;
		if (ret < 0) {
			PARA_ERROR_LOG("%s\n", strerror(errno));
			return -E_DCCP_WRITE;
		}
		written += ret;
	}
	return written;
}

static int queue_chunk_or_shutdown(struct dccp_client *dc, long unsigned chunk_num,
	size_t sent)
{
	int ret = cq_enqueue(dc->cq, chunk_num, sent);
	if (ret < 0) {
		PARA_NOTICE_LOG("enqueue error\n");
		dccp_shutdown_client(dc);
	}
	return ret;
}

static int send_queued_chunks(struct dccp_client *dc)
{
	struct queued_chunk *qc;
	while ((qc = cq_peek(dc->cq))) {
		char *buf;
		size_t len;
		int ret;
		cq_get(qc, &buf, &len);
		ret = dccp_write(dc->fd, buf, len);
		if (ret < 0)
			return ret;
		cq_update(dc->cq, ret);
		if (ret != len)
			return 1;
		cq_dequeue(dc->cq);
	}
	return 1;
}

static void dccp_send(long unsigned current_chunk,
		__a_unused long unsigned chunks_sent, const char *buf, size_t len)
{
	struct dccp_client *dc, *tmp;
	int ret;
	char *header_buf;
	size_t header_len;

	if (listen_fd < 0 || !len)
		return;

	list_for_each_entry_safe(dc, tmp, &clients, node) {
		if (!dc->header_sent && current_chunk) {
			header_buf = vss_get_header(&header_len);
			if (header_buf && header_len > 0) {
				if (queue_chunk_or_shutdown(dc, -1U, 0) < 0)
					continue;
			}
			dc->header_sent = 1;
		}
		ret = send_queued_chunks(dc);
		if (ret < 0) {
			dccp_shutdown_client(dc);
			continue;
		}
//		PARA_DEBUG_LOG("writing %d bytes to fd %d\n", len, dc->fd);
		ret = dccp_write(dc->fd, buf, len);
		if (ret < 0) {
			PARA_NOTICE_LOG("%s\n", PARA_STRERROR(-ret));
			dccp_shutdown_client(dc);
			continue;
		}
		if (ret != len)
			queue_chunk_or_shutdown(dc, current_chunk, ret);
	}
}

static void dccp_shutdown_clients(void)
{
	struct dccp_client *dc, *tmp;

	list_for_each_entry_safe(dc, tmp, &clients, node)
		dccp_shutdown_client(dc);
}

static char *dccp_info(void)
{
	static char *buf;
	int num_clients = 0;
	struct dccp_client *dc, *tmp;

	free(buf);
	list_for_each_entry_safe(dc, tmp, &clients, node)
		num_clients++;
	buf = make_message("dccp connected clients: %d\n", num_clients);
	return buf;
}

static char *dccp_help(void)
{
	return make_message("no help available\n");
}

/**
 * the init function of the dccp sender
 *
 * \param s pointer to the dccp sender struct
 *
 * It initializes all function pointers of \a s and starts
 * listening on the given port.
 */
void dccp_send_init(struct sender *s)
{
	int ret;

	INIT_LIST_HEAD(&clients);
	s->info = dccp_info;
	s->send = dccp_send;
	s->pre_select = dccp_pre_select;
	s->post_select = dccp_post_select;
	s->shutdown_clients = dccp_shutdown_clients;
	s->help = dccp_help;
	s->client_cmds[SENDER_ON] = NULL;
	s->client_cmds[SENDER_OFF] = NULL;
	s->client_cmds[SENDER_DENY] = NULL;
	s->client_cmds[SENDER_ALLOW] = NULL;
	s->client_cmds[SENDER_ADD] = NULL;
	s->client_cmds[SENDER_DELETE] = NULL;
	self = s;
	ret = dccp_open();
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
		s->status = SENDER_OFF;
	} else
		s->status = SENDER_ON;
}
